// requiem version 1.7.3.

#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <unistd.h>
#include <algorithm>
#include "md5.h"
#include "aes.h"
#include "sha1.h"

#if WINDOWS
#include <windows.h>
#endif

typedef unsigned char byte;

typedef long long int64;

#define error(args...) { fprintf(stderr, "ERROR: "); fprintf(stderr, ## args); exit(1); }
#define warning(args...) { fprintf(stderr, "WARNING: "); fprintf(stderr, ## args); }

bool verbose = false;
bool dump = false;

// from autogenerated tables.cc
extern byte table_carrot[];
extern byte table_garlic[];
extern byte table_potato[];
extern byte table_radish[];
extern byte table_tomato[];
extern byte table_metakey_array1[];

void copy(const void *src, void *dst, int bytes) {
  memcpy(dst, src, bytes);
}

int seek(FILE *f, int64 offset) {
#if WINDOWS
  //return _fseeki64(f, offset, SEEK_SET);
  // ugh - _fseeki64 isn't universal yet, so we have to
  // seek on the underlying file descriptor.
  fseek(f, 0, SEEK_SET); // clear buffers (HACK!!!)
  return _lseeki64(fileno(f), offset, SEEK_SET) == offset ? 0 : 1;
#else
  return fseeko(f, offset, SEEK_SET);
#endif
}

byte *read_file(char *file, int *size) {
  struct stat stat_buf;
  if (stat(file, &stat_buf)) {
    perror(NULL);
    error("can't open file %s\n", file);
  }
  int len = stat_buf.st_size;
  byte *data = (byte*)malloc(len);
  FILE *f = fopen(file, "rb");
  if (!f) error("can't open file %s\n", file);
  fread(data, 1, len, f);
  fclose(f);
  *size = len;
  return data;
}

void print_buf(char *name, byte *buf, int len) {
  if (dump) {
    FILE *f = fopen(name, "wb");
    fwrite(buf, 1, len, f);
    fclose(f);
  }
  if (!verbose) return;
  printf("%s\n", name);
  for (int i = 0; i < len; i++) {
    if ((i&0xf) == 0) printf("%08x: ", i);
    printf("%2x ", buf[i]);
    if ((i&0xf) == 0xf) {
      printf("  ");
      for (int j = i - 15; j <= i; j++) {
        printf("%c", buf[j] < 0x20 || buf[j] >= 0x7f ? ' ' : buf[j]);
      }
      printf("\n");
    }
  }
  if (len & 0xf) {
    for (int i = len & 0xf; i < 16; i++) printf("   ");
    printf("  ");
    for (int i = len & ~0xf; i < len; i++) {
      printf("%c", buf[i] < 0x20 || buf[i] >= 0x7f ? ' ' : buf[i]);
    }
    printf("\n");
  }
}

void twiddle(byte *buf, int len, int a, int b) {
  for (int i = 0; i < len; i++) {
    buf[i] = a*buf[i]+b;
  }
}

// parse a 4-byte big-endian number.
int get_int32(byte *p) {
  return (p[0]<<24) + (p[1]<<16) + (p[2]<<8) + p[3];
}
// store a 4-byte big-endian number.
void set_int32(byte *p, int v) {
  p[0] = v >> 24;
  p[1] = v >> 16;
  p[2] = v >> 8;
  p[3] = v;
}
// parse an 8-byte big-endian number.
int64 get_int64(byte *p) {
  unsigned int a = get_int32(p);
  unsigned int b = get_int32(p + 4);
  return ((int64)a << 32) | (int64)b;
}

// find the given atom in an m4p file.
// "p" should point to a list of atoms with total byte length "len".
byte *find_atom(byte *p, int len, char *types) {
  while (len >= 8) {
    int size = get_int32(p);
    if (size == 0) error("bad atom parsing\n");
    if (!strncmp((char*)p + 4, (char*)types, 4)) {
      if (!types[4]) return p;
      int children_offset = 8;
      if (!strncmp((char*)p + 4, "stsd", 4)) children_offset = 16;
      if (!strncmp((char*)p + 4, "drms", 4)) children_offset = 36;
      if (!strncmp((char*)p + 4, "drmi", 4)) children_offset = 86;
      return find_atom(p + children_offset, size - children_offset,
                       types + 5);
    }
    p += size;
    len -= size;
  }
  return NULL;
}

byte mac_address[6];
void get_mac_address() {
#if OSX
  // We should really try all mac addresses we can find and see if any
  // of them work.
  system("ifconfig en0 | grep ether > /tmp/mac_address");
  int size;
  byte *mac_addr = read_file("/tmp/mac_address", &size);
  system("rm -f /tmp/mac_address");
  byte *p = (byte*) memchr(mac_addr, ':', size);
  if (!p) error("can't parse mac address %s\n", mac_addr);
  p -= 2;
  for (int i = 0; i < 6; i++) {
    char buf[3];
    buf[0] = p[0];
    buf[1] = p[1];
    buf[2] = 0;
    mac_address[i] = strtol(buf, NULL, 16);
    if (i < 5 && p[2] != ':') error("can't parse mac address %s\n", mac_addr);
    if (i == 5 && p[2] != ' ') error("can't parse mac address %s\n", mac_addr);
    p += 3;
  }
#elif WINDOWS
  // build up a machine identifier in lieu of a MAC address
  byte tmp[16];
  HKEY hkey;
  md5_state_t ctx;

  md5_state_t main_ctx;
  md5_init(&main_ctx);
  md5_append(&main_ctx, (byte*) "cache-controlEthernet", 21);
  
  DWORD volume_serial_number;
  if (GetVolumeInformation("C:\\", NULL, 0, &volume_serial_number, NULL, NULL, NULL, 0)) {
    print_buf("volume serial number", (byte*)&volume_serial_number, 4);
    md5_init(&ctx);
    md5_append(&ctx, (byte*)&volume_serial_number, 4); // endian issues?
    md5_finish(&ctx, tmp);
    md5_append(&main_ctx, tmp, 4);
  }
  
  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "HARDWARE\\DESCRIPTION\\System", 0, KEY_READ, &hkey) == ERROR_SUCCESS) {
    byte bios_version[1024];
    DWORD bios_version_size = 1024;
    if (RegQueryValueEx(hkey, "SystemBiosVersion", 0, NULL, (BYTE*)&bios_version, &bios_version_size) == ERROR_SUCCESS) {
      print_buf("bios version", bios_version, bios_version_size);
      md5_init(&ctx);
      md5_append(&ctx, (byte*)bios_version, bios_version_size);
      md5_finish(&ctx, tmp);
      md5_append(&main_ctx, tmp, 4);
    }
    RegCloseKey(hkey);
  }

  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0", 0, KEY_READ, &hkey) == ERROR_SUCCESS) {
    byte processor_name[1024];
    DWORD processor_name_size = 1024;
    if (RegQueryValueEx(hkey, "ProcessorNameString", 0, NULL, (BYTE*)&processor_name, &processor_name_size) == ERROR_SUCCESS) {
      print_buf("processor name", processor_name, processor_name_size);
      md5_init(&ctx);
      md5_append(&ctx, processor_name, processor_name_size);
      md5_finish(&ctx, tmp);
      md5_append(&main_ctx, tmp, 4);
    }
    RegCloseKey(hkey);
  }

  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion", 0, KEY_READ, &hkey) == ERROR_SUCCESS) {
    byte product_id[1024];
    DWORD product_id_size = 1024;
    if (RegQueryValueEx(hkey, "ProductId", 0, NULL, (BYTE*)&product_id, &product_id_size) == ERROR_SUCCESS) {
      print_buf("product id", product_id, product_id_size);
      md5_init(&ctx);
      md5_append(&ctx, product_id, product_id_size);
      md5_finish(&ctx, tmp);
      md5_append(&main_ctx, tmp, 4);
    }
    RegCloseKey(hkey);
  }

  md5_finish(&main_ctx, tmp);
  copy(tmp, mac_address, 6);
#else
#error UNKNOWN OS
#endif
  print_buf("mac_address", mac_address, 6);
}

byte *decrypted_sidb;
byte *decrypted_sidd;

void decrypt_sidb() {
#if OSX
  char *sidb_filename = "/Users/Shared/SC Info/SC Info.sidb";
  char *sidd_filename = "/Users/Shared/SC Info/SC Info.sidd";
#elif WINDOWS
  char sidb_filename[1024];
  char sidd_filename[1024];
  char *all_users_profile = getenv("ALLUSERSPROFILE");
  if (strstr(all_users_profile, "Program")) { // hack to detect Vista
    snprintf(sidb_filename, 1024, "%s\\Apple Computer\\iTunes\\SC Info\\SC Info.sidb", all_users_profile);
    snprintf(sidd_filename, 1024, "%s\\Apple Computer\\iTunes\\SC Info\\SC Info.sidd", all_users_profile);
  } else { // pre-Vista
    snprintf(sidb_filename, 1024, "%s\\Application Data\\Apple Computer\\iTunes\\SC Info\\SC Info.sidb", all_users_profile);
    snprintf(sidd_filename, 1024, "%s\\Application Data\\Apple Computer\\iTunes\\SC Info\\SC Info.sidd", all_users_profile);
  }
#else
#error UNKNOWN OS
#endif

  // compute key
  byte key[16];
  for (int i = 0; i < 16; i++) {
    int mac = mac_address[i % 6];
    int shr = (0x84db7605LL * (0x7769c46 + 0x3bb0*mac + 0x3bb000*i)) >> 47;
    key[i] = 0x46 + 0xb0*mac + 0x5c*shr;
  }
  print_buf("sidb_key", key, 16);
  
  // iv is a fixed constant in the code
  byte iv[16] = {0x2c, 0x67, 0xd5, 0xc1, 0x0c, 0xf4, 0x27, 0x3d,
                 0xd4, 0x06, 0xce, 0x0f, 0x8f, 0xd0, 0x42, 0xa6};
  print_buf("sidb_iv", iv, 16);
  
  int len;
  byte *sidb = read_file(sidb_filename, &len);
  if (get_int32(sidb) != 0x00060002) {
    fprintf(stderr, "WARNING: possible sidb format change %x\n",
            get_int32(sidb));
  }
  sidb += 4;
  len -= 4;
  
  decrypted_sidb = (byte*) malloc(len);
  aes_context aes_ctx;
  aes_set_key(&aes_ctx, key, 128);
  aes_cbc_decrypt(&aes_ctx, sidb, decrypted_sidb, len, iv);
  
  print_buf("sidb", decrypted_sidb, len);
  int padding = 0;
  while (!get_int32(decrypted_sidb)) {  // skip leading zeroes: is this right?
    decrypted_sidb += 4;
    padding += 4;
  }
  // what is this trailing junk for?
  //int n = get_int32(decrypted_sidb);
  //if (n + padding != len) error("bad length %d %d\n", n + padding, len);
  if (strncmp((char*)decrypted_sidb + 4, "sean", 4)) {
    error("bad decrypt of sidb\n");
  }

  byte *sidd = read_file(sidd_filename, &len);
  if (get_int32(sidd) != 0x00010001) {
    fprintf(stderr, "WARNING: possible sidd format change %x\n",
            get_int32(sidd));
  }
  sidd += 4;
  len -= 4;
  
  decrypted_sidd = (byte*) malloc(len);
  aes_cbc_decrypt(&aes_ctx, sidd, decrypted_sidd, len, iv);
  print_buf("sidd", decrypted_sidd, len);
  padding = 0;
  while (!get_int32(decrypted_sidd)) {  // skip leading zeroes: is this right?
    decrypted_sidd += 4;
    padding += 4;
  }
  // what is this trailing junk for?
  //int n = get_int32(decrypted_sidd);
  //if (n + padding != len) error("bad length %d %d\n", n + padding, len);
  if (strncmp((char*)decrypted_sidd + 4, "dbag", 4)) {
    error("bad decrypt of sidd\n");
  }
}

byte *find_in_sid(byte *p, char *types, int nchildren) {
  while (nchildren > 0) {
    if (!strncmp((char*)p + 4, (char*)types, 4)) {
      if (!types[4]) return p;
      nchildren = get_int32(p + 12);
      return find_in_sid(p + 20, types + 5, nchildren);
    }
    p += get_int32(p);
    nchildren--;
  }
  error("couldn't find %s\n", types);
}
byte *find_in_sidb(char *types) {
  return find_in_sid(decrypted_sidb, types, 1);
}
byte *find_in_sidd(char *types) {
  return find_in_sid(decrypted_sidd, types, 1);
}

typedef struct {
  int sign;                   // 1 or -1
  int n;                      // # of digits of base 2^32
  unsigned int digits[100];   // digits, little-endian order
} bigint;

void bigint_print(char *name, const bigint &a) {
  if (!verbose) return;
  printf("%8s = ", name);
  if (a.n == 0) { printf("0\n"); return; }
  if (a.sign == -1) printf("-");
  printf("%x", a.digits[a.n - 1]);
  for (int i = a.n - 2; i >= 0; i--) {
    printf("%08x", a.digits[i]);
  }
  printf("\n");
}

void bigint_zero(bigint *a) {
  a->sign = 1;
  a->n = 0;
}
void bigint_one(bigint *a) {
  a->sign = 1;
  a->n = 1;
  a->digits[0] = 1;
}

void bigint_normalize(bigint *a) {
  // truncate leading zeroes
  while (a->n > 0 && a->digits[a->n - 1] == 0) a->n--;
  
  // zero gets a positive sign
  if (a->n == 0) a->sign == 1;
}

// r = a + b
void bigint_add(bigint *r, const bigint &a, const bigint &b);
// r = a - b
void bigint_sub(bigint *r, const bigint &a, const bigint &b);
// returns a > b
bool bigint_gt(const bigint &a, const bigint &b);

void bigint_add(bigint *r, const bigint &a, const bigint &b) {
  bigint x = a;
  bigint y = b;
  
  if (x.sign == 1 && y.sign == -1) {
    y.sign = 1;
    bigint_sub(r, x, y);
    return;
  }
  if (x.sign == -1 && y.sign == 1) {
    x.sign = 1;
    bigint_sub(r, y, x);
    return;
  }
  if (x.sign == -1 && y.sign == -1) {
    x.sign = 1;
    y.sign = 1;
    bigint_add(r, x, y);
    r->sign = -1;
    return;
  }
  if (bigint_gt(y, x)) {
    bigint_add(r, y, x);
    return;
  }
  
  // do addition
  int carry = 0;
  for (int i = 0; i < x.n; i++) {
    unsigned int oldx = x.digits[i];
    unsigned int oldy = i < y.n ? y.digits[i] : 0;
    unsigned int newx = oldx + oldy + carry;
    x.digits[i] = newx;
    carry = newx < oldx || (carry == 1 && newx == oldx) ? 1 : 0;
  }
  
  if (carry) {
    x.digits[x.n++] = 1;
  }
  
  bigint_normalize(&x);
  *r = x;
}

void bigint_sub(bigint *r, const bigint &a, const bigint &b) {
  bigint x = a;
  bigint y = b;
  
  if (x.sign == 1 && y.sign == -1) {
    y.sign = 1;
    bigint_add(r, x, y);
    return;
  }
  if (x.sign == -1 && y.sign == 1) {
    x.sign = 1;
    bigint_add(r, x, y);
    r->sign *= -1;
    return;
  }
  if (x.sign == -1 && y.sign == -1) {
    x.sign = 1;
    y.sign = 1;
    bigint_sub(r, y, x);
    return;
  }
  if (bigint_gt(y, x)) {
    bigint_sub(r, y, x);
    r->sign *= -1;
    return;
  }
  
  // do subtraction
  int borrow = 0;
  for (int i = 0; i < x.n; i++) {
    unsigned int oldx = x.digits[i];
    unsigned int oldy = i < y.n ? y.digits[i] : 0;
    x.digits[i] = oldx - oldy - borrow;
    borrow = oldx < oldy || (borrow == 1 && oldx == oldy) ? 1 : 0;
  }
  
  // should never happen since x >= y.
  if (borrow) error("borrowing out of top!\n");
  
  bigint_normalize(&x);
  *r = x;
}

void bigint_mul(bigint *r, const bigint &a, const bigint &b) {
  bigint sum;
  bigint_zero(&sum);
  
  for (int i = 0; i < a.n; i++) {
    for (int j = 0; j < b.n; j++) {
      unsigned long long t = ((unsigned long long)a.digits[i]) * b.digits[j];
      bigint term;
      term.sign = 1;
      term.n = i + j + 2;
      for (int k = 0; k < i+j; k++) term.digits[k] = 0;
      term.digits[i+j] = t;
      term.digits[i+j+1] = t>>32;
      bigint_add(&sum, sum, term);
    }
  }
  sum.sign = a.sign * b.sign;
  *r = sum;
}


void bigint_shiftR(bigint *r, const bigint &a, int shift) {
  bigint res;
  int chunks = shift / 32;
  int offset = shift % 32;
  for (int i = 0; i < a.n; i++) {
    unsigned long long x = a.digits[i];
    unsigned long long y = i+1 < a.n ? a.digits[i+1] : 0;
    unsigned long long z = x | (y << 32);
    res.digits[i - chunks] = z >> offset;
  }
  res.sign = a.sign;  // this shift is round-toward-zero semantics
  res.n = a.n - chunks;
  bigint_normalize(&res);
  *r = res;
}

void bigint_shiftL(bigint *a, int shift) {
  int chunks = shift / 32;
  int offset = shift % 32;
  if (offset) error("can't do offset shifts yet %d\n", shift);
  for (int i = a->n - 1; i >= 0; i--) {
    a->digits[i + chunks] = a->digits[i];
  }
  for (int i = 0; i < chunks; i++) a->digits[i] = 0;
  a->n += chunks;
}

// returns true iff a>b
bool bigint_gt(const bigint &a, const bigint &b) {
  if (a.sign == 1 && b.sign == -1) return true;
  if (a.sign == -1 && b.sign == 1) return false;
  if (a.sign == -1 && b.sign == -1) {
    bigint x = a;
    bigint y = b;
    x.sign = 1;
    y.sign = 1;
    return bigint_gt(y, x);
  }
  
  if (a.n > b.n) return true;
  if (a.n < b.n) return false;
  for (int i = a.n - 1; i >= 0; i--) {
    if (a.digits[i] > b.digits[i]) return true;
    if (a.digits[i] < b.digits[i]) return false;
  }
  return false;  // a == b
}

// requires a>0
int bigint_nbits(const bigint &a) {
  int n = 32 * a.n - 32;
  unsigned int z = a.digits[a.n - 1];
  while (z != 0) {
    n++;
    z >>= 1;
  }
  return n;
}

bool bigint_bit(const bigint &a, int b) {
  int digit = b / 32;
  int offset = b % 32;
  return ((a.digits[digit] >> offset) & 1) != 0;
}

void build_bigint(bigint *r, byte* buf, int n) {
  r->sign = 1;
  r->n = (n + 3) / 4;
  int accum = 0;
  for (int i = 0; i < n; i++) {
    if ((i&3) == 0) accum = 0;
    int z = buf[n - 1 - i];
    
    int or46 = 0xff|(0x307e + 0xfffffa29*z);
    int xor47 = (or46)^(0x2327b048 + 0xfff38978*z);
    int or48 = (0x307e + 0xfffffa29*z)|(0xdcd84fb7 + 0xc7688*z);
    int or49 = 0xff|(or48);
    int shl82 = (0x1)<<((i&3)*8);
    int shl86 = (0xdcd88033 + 0xffffffff*xor47 + 0xfffffffe*or49 + 0xc70b1*z)<<((i&3)*8);
    int or95 = (0xffffffff + 0x1*shl82 + 0x1*shl86)|(-1 - accum);
    
    accum += 1 - shl82 - shl86 + or95;
    r->digits[i/4] = accum;  // writes the same loc up to 4 times.
  }
  bigint_normalize(r);
}

// get the key used to decrypt the keys stored in the sidb file.
void get_metakey(byte *metakey, int user_id, int key_id) {
  
  // starting with mac address, user id, and key id
  byte metakey_starter[16];
  copy(mac_address, metakey_starter, 6);
  set_int32(metakey_starter + 6, key_id);
  set_int32(metakey_starter + 10, user_id);
  metakey_starter[14] = 0xff;
  metakey_starter[15] = 0xff;
  print_buf("metakey starter", metakey_starter, 16);
  
  // get array1 table
  byte *array1 = table_metakey_array1;
  
  // get a table out of the sidd file.
  // check version number
  if (get_int32(find_in_sidd("dbag.vers") + 20) != 0x10001) {
    warning("bad sean.vers %x\n", get_int32(find_in_sidd("dbag.vers") + 20));
  }

  // find the right group (a.k.a. mac address, a.k.a. machine)
  while (1) {
    byte *guid = find_in_sidd("dbag.grup.head.guid");
    if (memcmp(guid + 20, mac_address, 6)) {
      copy("----", find_in_sidd("dbag.grup") + 4, 4);  // erase group
      continue;
    }
    break;
  }
  
  // find the right user
  while (1) {
    byte *user = find_in_sidd("dbag.grup.user");
    if (get_int32(user + 8) != user_id) {
      copy("----", user + 4, 4); // erase user
      continue;
    }
    break;
  }
  
  // find the usag entry
  while (1) {
    byte *usag = find_in_sidd("dbag.grup.user.usag");
    if (get_int32(usag + 8) != 0x2100003) {
      copy("----", usag + 4, 4); // erase usag
      continue;
    }
    break;
  }
  
  // grab data we need (table data & algorithm)
  byte *valu = find_in_sidd("dbag.grup.user.usag.valu");
  byte *table = valu + 0x286;
  char *algr = (char*)valu + 0x24;
  
  // munge with table lookups, encryption-ish
  byte pre_buf[16];
  copy(metakey_starter, pre_buf, 16);
  for (int round = 0; round < 9; round++) {
    int scramble[16];
    for (int i = 0; i < 16; i++) {
      scramble[i] = get_int32(table + 16384*round + 1024*i +
                              4*pre_buf[i*5&0xf]) * 0x50a19981 + 0x5e39c9c0;
    }
    for (int i = 0; i < 4; i++) {
      int scramble0 = scramble[4*i];
      int scramble4 = scramble[4*i+1];
      int scramble8 = scramble[4*i+2];
      int scramble12 = scramble[4*i+3];
      int or0 = (0x6cd8d601 + 0xa2ed4502*scramble8) | (0x6cd8d601 + 0xa2ed4502*scramble4);
      int or1 = (0xd9b1ac01 + 0xa2ed4502*scramble4 + 0xa2ed4502*scramble8 + 0xfffffffe*or0) | (0x6cd8d601 + 0xa2ed4502*scramble12);
      int or2 = (0x468a8201 + 0xa2ed4502*scramble12 + 0xfffffffe*or0 + 0xa2ed4502*scramble4 + 0xa2ed4502*scramble8 + 0xfffffffe*or1) | (0x6cd8d601 + 0xa2ed4502*scramble0);
      int t = 0x8dc2c100 + 0x1*scramble4 + 0x1*scramble0 + 0xd3f0627f*or2 + 0x1*scramble8 + 0xd3f0627f*or1 + 0x1*scramble12 + 0xd3f0627f*or0;
      t = 0xc99394ff + 0x2e895d7f*t;
      for (int j = 0; j < 4; j++) {
        pre_buf[4*i+j] = t >> (24-8*j);
      }
    }
  }
  
  // get your scramble-tude on...
  byte *p = table + 4096*9*4 + 4;
  byte *q = array1;
  byte t;
  pre_buf[ 0] = q[ 2*256 + p[         pre_buf[ 0]]]*0x99+0xdf;
  pre_buf[ 4] = q[ 3*256 + p[ 4*256 + pre_buf[ 4]]]*0xb5+0xff;
  pre_buf[ 8] = q[ 4*256 + p[ 8*256 + pre_buf[ 8]]]*0x59+0xd3;
  pre_buf[12] = q[ 5*256 + p[12*256 + pre_buf[12]]]*0x2f+0x1e;
  t = pre_buf[1];
  pre_buf[ 1] = q[ 6*256 + p[ 1*256 + pre_buf[ 5]]]*0x95+0xfc;
  pre_buf[ 5] = q[ 7*256 + p[ 5*256 + pre_buf[ 9]]]*0x05+0x36;
  pre_buf[ 9] = q[ 0*256 + p[ 9*256 + pre_buf[13]]]*0x05+0x6d;
  pre_buf[13] = q[ 9*256 + p[13*256 + t]]*0x05+0xe1;
  t = pre_buf[2];
  pre_buf[ 2] = q[10*256 + p[ 2*256 + pre_buf[10]]]*0x4d+0xb0;
  pre_buf[10] = q[11*256 + p[10*256 + t]]*0x0b+0x2e;
  t = pre_buf[6];
  pre_buf[ 6] = q[12*256 + p[ 6*256 + pre_buf[14]]]*0x93+0x4e;
  pre_buf[14] = q[13*256 + p[14*256 + t]]*0x9b+0x75;
  t = pre_buf[3];
  pre_buf[ 3] = q[14*256 + p[ 3*256 + pre_buf[15]]]*0x93+0xc8;
  pre_buf[15] = q[15*256 + p[15*256 + pre_buf[11]]]*0x33+0x8b;
  pre_buf[11] = q[ 1*256 + p[11*256 + pre_buf[ 7]]]*0x5b+0x25;
  pre_buf[ 7] = q[ 8*256 + p[ 7*256 + t]]*0xed+0xc5;

  print_buf("common_alg", pre_buf, 16);

  char output_algr[7];
  copy(algr, output_algr, 6);
  output_algr[6] = 0;
  if (verbose) printf("using algorithm \'%s\'\n", output_algr);
  
  // figure out modifications needed for each algorithm
  if (!strncmp(algr, "carrot", 6)) {
    byte *p = table_carrot;
    for (int i = 0; i < 16; i++) pre_buf[i] = p[pre_buf[i]] ^ 0x5d;
    twiddle(pre_buf, 16, 113, 28);
  } else if (!strncmp(algr, "celery", 6)) {
    twiddle(pre_buf, 16, 235, 135);
  } else if (!strncmp(algr, "fennel", 6)) {
    twiddle(pre_buf, 16, 197, 217);
  } else if (!strncmp(algr, "garlic", 6)) {
    byte *p = table_garlic;
    for (int i = 0; i < 16; i++) pre_buf[i] = p[pre_buf[i]] ^ 0x5d;
    twiddle(pre_buf, 16, 197, 240);
  } else if (!strncmp(algr, "peanut", 6)) {
    twiddle(pre_buf, 16, 103, 109);
  } else if (!strncmp(algr, "potato", 6)) {
    byte *p = table_potato;
    for (int i = 0; i < 16; i++) pre_buf[i] = p[pre_buf[i]] ^ 0x5d;
    twiddle(pre_buf, 16, 55, 250);
  } else if (!strncmp(algr, "radish", 6)) {
    byte *p = table_radish;
    for (int i = 0; i < 16; i++) pre_buf[i] = p[pre_buf[i]] ^ 0x5d;
    twiddle(pre_buf, 16, 155, 70);
  } else if (!strncmp(algr, "squash", 6)) {
    twiddle(pre_buf, 16, 249, 74);
  } else if (!strncmp(algr, "tomato", 6)) {
    byte *p = table_tomato;
    for (int i = 0; i < 16; i++) pre_buf[i] = p[pre_buf[i]] ^ 0x5d;
    twiddle(pre_buf, 16, 23, 239);
  } else if (!strncmp(algr, "turnip", 6)) {
    twiddle(pre_buf, 16, 239, 4);
  } else {
    error("unknown algorithm \'%s\'\n", output_algr);
  }
  
  print_buf("pre_buf", pre_buf, 16);
  
  bigint x;
  {
    SHA1Context sha;
    SHA1Reset(&sha);
    SHA1Input(&sha, metakey_starter, 16);
    SHA1Result(&sha);
    byte buf[20];
    for (int i = 0; i < 5; i++) set_int32(buf + 4*i, sha.Message_Digest[i]);
    twiddle(buf, 20, 231, 153);
    for (int i = 0; i < 14; i++) {
      byte x = pre_buf[i+2];
      byte y = buf[i+1];
      int or27 = 0xff|(0xa07e + 0xffff9129*x);
      int xor29 = (0x109abd82 + 0xffad8735*x)^(or27);
      int or31 = (0xa07e + 0xffff9129*x)|(0xef65427d + 0x5278cb*x);
      int or32 = 0xff|(or31);
      int or65 = 0xff|(0x417e + 0xffffcb29*y);
      int or68 = (0x417e + 0xffffcb29*y)|(0xd597c97d + 0x698ac6*y);
      int or69 = 0xff|(or68);
      int xor75 = (or65)^(0x2a683682 + 0xff96753a*y);
      int or97 = (0xab3015f3 + 0xfffffffe*xor75 + 0xd2abde*y + 0xfffffffc*or69)|(0xdecbc5f3 + 0xa413e8*x + 0xfffffffe*xor29 + 0xfffffffc*or32);
      buf[i+1] = 0xb7abcb54 + 0xe20e1dce*or32 + 0x71070ee7*xor29 + 0x71070ee7*xor75 + 0x5b76357*y + 0x3fc1acd4*x + 0xe20e1dce*or69 + 0x71070ee7*or97;
    }
    int m = mac_address[4];
    int or24 = 0xff|(0xb8ff + 0xffff9cff*m);
    int or26 = (0xb8ff + 0xffff9cff*m)|(0x21e2d7ab + 0x32fa52*m);
    int xor28 = (or24)^(0xde1d2854 + 0xffcd05ae*m);
    int or32 = 0xff|(or26);
    int and41 = 0x80000003&(0xde1c6f57 + 0x1*xor28 + 0xffcd68af*m + 0x2*or32);
    int n = 8 + and41;
    build_bigint(&x, buf, n);
  }
  bigint_print("x", x);
  
  bigint y;
  build_bigint(&y, pre_buf, 16);
  bigint_print("y", y);
  
  // some constant from the QTC library
  byte pre_zbuf[20] = {0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x8a, 0x4c, 0xcd, 0x5b, 0x5b, 0x5b, 0x5b, 0x5b, 0x7a, 0x5c, 0x6f, 0x8b, 0x7e};
  for (int i = 0; i < 20; i++) pre_zbuf[i] ^= 0xc2;
  
  bigint z;
  build_bigint(&z, pre_zbuf, 20);
  bigint_print("z", z);
  
  // computes d = floor(2^320/z).
  bigint d;
  bigint_one(&d);
  bigint_shiftL(&d, 160);
  while (true) {
    bigint c = d;
    bigint a;
    bigint_mul(&a, d, d);
    bigint_shiftR(&a, a, 160);
    bigint_mul(&a, a, z);
    bigint_shiftR(&a, a, 160);
    bigint_add(&d, d, d);
    bigint_sub(&d, d, a);
    if (!bigint_gt(d, c)) break;
  }
  {
    bigint a;
    bigint_one(&a);
    bigint_shiftL(&a, 320);
    bigint c;
    bigint_mul(&c, d, z);
    bigint_sub(&a, a, c);
    bigint_one(&c);
    bigint_sub(&d, d, c);
    bigint_add(&a, a, z);
  }
  bigint_print("2^320/z", d);
  // z, d constants - don't use the above code, just materialize them here.
  
  // compute w
  bigint w;
  bigint_one(&w);
  for (int i = bigint_nbits(x) - 1; i >= 0; i--) {
    bigint_mul(&w, w, w);
    bigint_mul(&y, y, y);
    if (bigint_bit(x, i)) {
      bigint_sub(&w, y, w);
      bigint c;
      bigint_shiftR(&c, w, 159);
      bigint_mul(&c, c, d);
      bigint_shiftR(&c, c, 161);
      bigint_mul(&c, c, z);
      bigint_sub(&w, w, c);
      while (w.sign == -1) bigint_add(&w, w, z);
      while (!bigint_gt(z, w)) bigint_sub(&w, w, z);
      bigint_shiftR(&c, y, 159);
      bigint_mul(&c, c, d);
      bigint_shiftR(&c, c, 161);
      bigint_mul(&c, c, z);
      bigint_sub(&y, y, c);
      while (y.sign == -1) bigint_add(&y, y, z);
      while (!bigint_gt(z, y)) bigint_sub(&y, y, z);
    } else {
      bigint_sub(&y, y, w);
      bigint c;
      bigint_shiftR(&c, y, 159);
      bigint_mul(&c, c, d);
      bigint_shiftR(&c, c, 161);
      bigint_mul(&c, c, z);
      bigint_sub(&y, y, c);
      while (y.sign == -1) bigint_add(&y, y, z);
      while (!bigint_gt(z, y)) bigint_sub(&y, y, z);
      bigint_shiftR(&c, w, 159);
      bigint_mul(&c, c, d);
      bigint_shiftR(&c, c, 161);
      bigint_mul(&c, c, z);
      bigint_sub(&w, w, c);
      while (w.sign == -1) bigint_add(&w, w, z);
      while (!bigint_gt(z, w)) bigint_sub(&w, w, z);
    }
  }
  bigint_print("w", w);
  
  byte sha_arg[32];
  for (int i = 0; i < w.n; i++) {
    set_int32(sha_arg + 28 - 4*i, w.digits[i]);
  }
  for (int i = w.n; i < 8; i++) {
    set_int32(sha_arg + 28 - 4*i, 0);
  }
  print_buf("sha_arg", sha_arg, 32);
  
  {
    SHA1Context sha;
    SHA1Reset(&sha);
    SHA1Input(&sha, sha_arg, 32);
    SHA1Result(&sha);
    byte buf[20];
    for (int i = 0; i < 5; i++) set_int32(buf + 4*i, sha.Message_Digest[i]);
    copy(buf + 1, metakey, 16);
  }
  print_buf("meta_key", metakey, 16);
}

byte *decrypt_priv(byte *priv, int user_id, int key_id,
                   byte *name, byte *iviv) {
  // get metakey to decrypt the key from the sidb file.
  byte metakey[16];
  get_metakey(metakey, user_id, key_id);
  
  // check version number
  if (get_int32(find_in_sidb("sean.vers") + 20) != 0x10001) {
    warning("bad sean.vers %x\n", get_int32(find_in_sidb("sean.vers") + 20));
  }
  
  // find the right group (a.k.a. mac address, a.k.a. machine)
  while (1) {
    byte *guid = find_in_sidb("sean.grup.head.guid");
    if (memcmp(guid + 20, mac_address, 6)) {
      copy("----", find_in_sidb("sean.grup") + 4, 4);  // erase group
      continue;
    }
    break;
  }
  
  // find the right user
  while (1) {
    byte *user = find_in_sidb("sean.grup.user");
    if (get_int32(user + 8) != user_id) {
      copy("----", user + 4, 4); // erase user
      continue;
    }
    // check user version
    byte *vers = find_in_sidb("sean.grup.user.vers");
    if (get_int32(vers + 20) != 0x10002) {
      warning("bad sean.grup.user.vers %x\n", get_int32(vers + 20));
    }
    break;
  }
  
  // find the right key
  while (1) {
    byte *key_atom = find_in_sidb("sean.grup.user.key ");
    if (get_int32(key_atom + 8) != key_id) {
      copy("----", key_atom + 4, 4); // erase key
      continue;
    }
    // check key type
    byte *type = find_in_sidb("sean.grup.user.key .type");
    if (get_int32(type + 20) != 3) {
      // NOTE: Type 0 used in old style sidb files.  If you see type 0,
      //       requiem 1.3 should be able to deDRM your file.
      //       Type 2 used in .m4r (ringtones)
      warning("bad sean.grup.user.key .type %d\n", get_int32(type + 20));
    }
    break;
  }
  
  // extract encrypted key
  byte encrypted_privkey[16];
  copy(find_in_sidb("sean.grup.user.key .valu") + 20, encrypted_privkey, 16);
  print_buf("encrypted priv_key", encrypted_privkey, 16);
  
  // decrypt the key from the sidb file using the metakey
  aes_context aes_ctx;
  aes_set_key(&aes_ctx, metakey, 128);
  byte key[16];
  aes_decrypt(&aes_ctx, encrypted_privkey, key);
  print_buf("priv_key", key, 16);
  
  // compute iv
  byte iv[16];
  {
    // compute md5 of concatenation of name and iviv
    md5_state_t ctx;
    md5_init(&ctx);
    md5_append(&ctx, name + 8, strlen((char*)name + 8));
    md5_append(&ctx, iviv + 8, 16);
    md5_finish(&ctx, iv);
  }
  print_buf("priv_iv", iv, 16);
  
  // decrypt priv atom with key and iv we calculated.
  int len = get_int32(priv);
  byte *decrypted_priv = (byte*) malloc(len);
  copy(priv, decrypted_priv, 8);
  aes_set_key(&aes_ctx, key, 128);
  aes_cbc_decrypt(&aes_ctx, priv + 8, decrypted_priv + 8, len - 8, iv);
  print_buf("priv", decrypted_priv, len);
  if (strncmp((char*)decrypted_priv + 8, "itun", 4)) {
    error("bad decrypt of priv\n");
  }
  return decrypted_priv;
}

byte *find_in_priv(byte *decrypted_priv, char *type) {
  byte *p = decrypted_priv + 12;
  int n = get_int32(decrypted_priv) - 12;
  
  while (n >= 8) {
    if (!strncmp((char*)p + 4, type, 4)) {
      return p;
    }
    int m = get_int32(p);
    p += m;
    n -= m;
  }
  return NULL;
}

byte *decrypt_priv_in_trak(byte *trak) {
  int n = get_int32(trak);
  
  // find drm atom
  byte *drms;
  byte *drms_body;
  if (find_atom(trak + 8, n - 8, "mdia.minf.stbl.stsd.drms")) {
    drms = find_atom(trak + 8, n - 8, "mdia.minf.stbl.stsd.drms");
    drms_body = drms + 36;
  } else if (find_atom(trak + 8, n - 8, "mdia.minf.stbl.stsd.drmi")) {
    drms = find_atom(trak + 8, n - 8, "mdia.minf.stbl.stsd.drmi");
    drms_body = drms + 86;
  } else {
    // no DRM found in this trak.
    return NULL;
  }
  
  int m = get_int32(drms) - (drms_body - drms);
  byte *sinf = find_atom(drms_body, m, "sinf");
  byte *frma = find_atom(drms_body, m, "sinf.frma");
  byte *priv = find_atom(drms_body, m, "sinf.schi.priv");
  byte *user = find_atom(drms_body, m, "sinf.schi.user");
  byte *key_ = find_atom(drms_body, m, "sinf.schi.key ");
  byte *name = find_atom(drms_body, m, "sinf.schi.name");
  byte *iviv = find_atom(drms_body, m, "sinf.schi.iviv");
  int user_id = get_int32(user + 8);
  int key_id = get_int32(key_ + 8);
  
  // NOTE: if the user_id and key_id are different for different
  // drms's in the file, the priv decrypter might not find
  // the decryption key because it does a destructive lookup
  // in the sidb file.  Need to fix?
  
  return decrypt_priv(priv, user_id, key_id, name, iviv);
}

void decrypt_trak(FILE *f, FILE *g, byte *trak, int64 trak_offset) {
  byte *decrypted_priv = decrypt_priv_in_trak(trak);
  if (decrypted_priv == NULL) return;  // no DRM found in this trak.
  
  // find drm atom
  int n = get_int32(trak);
  byte *drms;
  byte *drms_body;
  if (find_atom(trak + 8, n - 8, "mdia.minf.stbl.stsd.drms")) {
    drms = find_atom(trak + 8, n - 8, "mdia.minf.stbl.stsd.drms");
    drms_body = drms + 36;
  } else if (find_atom(trak + 8, n - 8, "mdia.minf.stbl.stsd.drmi")) {
    drms = find_atom(trak + 8, n - 8, "mdia.minf.stbl.stsd.drmi");
    drms_body = drms + 86;
  } else {
    error("drmx not found\n");
  }
  
  int m = get_int32(drms) - (drms_body - drms);
  byte *sinf = find_atom(drms_body, m, "sinf");
  byte *frma = find_atom(drms_body, m, "sinf.frma");
  byte *skcr = find_atom(drms_body, m, "sinf.skcr");
  
  // grab key & iv from priv
  byte *k = find_in_priv(decrypted_priv, "key ");
  if (!k) error("can't find key in priv\n");
  if (get_int32(k) != 0x18) error("key in priv has bad size\n");
  byte *i = find_in_priv(decrypted_priv, "iviv");
  if (!i) error("can't find iviv in priv\n");
  if (get_int32(i) != 0x18) error("iviv in priv has bad size\n");
  byte key[16];
  byte iv[16];
  copy(k + 8, key, 16);
  copy(i + 8, iv, 16);
  
  // neuter the sinf atom by changing it to a free atom.
  seek(g, trak_offset + (sinf - trak) + 4);
  fwrite("free", 1, 4, g);
  
  // frma contains the new name for drms
  seek(g, trak_offset + (drms - trak) + 4);
  fwrite(frma + 8, 1, 4, g);
  
  // tables that describe where the trak samples are stored.
  byte *stsc = find_atom(trak + 8, n - 8, "mdia.minf.stbl.stsc");
  byte *stsz = find_atom(trak + 8, n - 8, "mdia.minf.stbl.stsz");
  byte *stco = find_atom(trak + 8, n - 8, "mdia.minf.stbl.stco");
  byte *co64 = find_atom(trak + 8, n - 8, "mdia.minf.stbl.co64");
  
  // don't bother handling constant-sized samples (never happens?)
  if (get_int32(stsz + 12) != 0) error("bad stsz - constant sized!\n");
  
  int print_blocks = verbose ? 1 : 0;
  
  // set up aes context
  aes_context aes_ctx;
  aes_set_key(&aes_ctx, key, 128);
  
  // Loop through chunk groups, chunks, and samples, decrypting each
  // sample we find.
  int sample = 0;
  int chunk_groups = get_int32(stsc + 12);
  for (int chunk_group = 0; chunk_group < chunk_groups; chunk_group++) {
    int first_chunk = get_int32(stsc + 16 + 12 * chunk_group) - 1;
    int samples_per_chunk = get_int32(stsc + 16 + 12 * chunk_group + 4);
    int chunks;
    if (chunk_group != chunk_groups - 1) {
      chunks = get_int32(stsc + 16 + 12 * (chunk_group + 1)) - 1 - first_chunk;
    } else {
      chunks = get_int32(stco + 12) - first_chunk;
    }
    for (int chunk = first_chunk; chunk < first_chunk + chunks; chunk++) {
      int64 sample_offset;
      if (stco) {
        sample_offset = get_int32(stco + 16 + 4 * chunk);
      } else {
        sample_offset = get_int64(co64 + 16 + 8 * chunk);
      }
      
      // Loop over samples in chunk and decrypt them.  Each sample is
      // encrypted individually.
      for (int i = 0; i < samples_per_chunk; i++, sample++) {
        int sample_size = get_int32(stsz + 20 + sample * 4);
        byte *ciphertext = (byte*) malloc(sample_size);
        byte *plaintext = (byte*) malloc(sample_size);
        seek(f, sample_offset);
        fread(ciphertext, 1, sample_size, f);
        if (print_blocks) print_buf("ciphertext", ciphertext, sample_size);
        if (skcr) {
          int init = get_int32(skcr + 8);
          int crypt_size = get_int32(skcr + 12);
          int uncrypt_size = get_int32(skcr + 16);
          byte *local_iv = iv;
          
          // initial unencrypted section
          int n = std::min(sample_size, init);
          copy(ciphertext, plaintext, n);
          
          for (int j = n; j < sample_size; ) {
            // decrypt encrypted section
            n = std::min(sample_size - j, crypt_size);
            aes_cbc_decrypt(&aes_ctx, ciphertext + j, plaintext + j, n, local_iv);
            j += n;
            local_iv = ciphertext + j - 16;  // chain iv
            
            // copy over unencrypted section
            n = std::min(sample_size - j, uncrypt_size);
            copy(ciphertext + j, plaintext + j, n);
            j += n;
          }
        } else {
          int full_block_size = sample_size & ~0xF;
          aes_cbc_decrypt(&aes_ctx, ciphertext, plaintext, full_block_size, iv);
          copy(ciphertext + full_block_size, plaintext + full_block_size,
               sample_size - full_block_size);
        }
        if (print_blocks) {
          print_buf("plaintext", plaintext, sample_size);
          print_blocks--;
        }
        seek(g, sample_offset);
        fwrite(plaintext, 1, sample_size, g);
        free(ciphertext);
        free(plaintext);
        sample_offset += sample_size;
      }
    }
  }
}

int main(int argc, char *argv[]) {
  int c;
  while ((c = getopt (argc, argv, "vd")) != -1) {
    switch (c) {
      case 'v': verbose = 1; break;
      case 'd': dump = 1; break;
      default: error("bad option %c\n", c);
    }
  }
  
  char *input_file = argv[optind];
  char *output_file = argv[optind + 1];

  byte buf[4096];
  
  // do some setup
  get_mac_address();
  decrypt_sidb();

  // find size of input file
  struct stat stat_buf;
  if (stat(input_file, &stat_buf)) {
    perror(NULL);
    error("can't open file %s\n", input_file);
  }
  int len = stat_buf.st_size;

  // find and load the moov atom (the metadata) into memory.
  FILE *f = fopen(input_file, "rb");
  int64 moov_offset = 0;
  int moov_len = 0;
  while (1) {
    if (seek(f, moov_offset)) error("bad seek - moov missing?\n");
    if (fread(buf, 1, 8, f) != 8) error("bad read - moov missing?\n");
    moov_len = get_int32(buf);
    if (moov_len == 1) moov_len = get_int64(buf + 8);
    if (moov_len == 0) error("can't find moov atom\n");
    if (!strncmp((char*)buf + 4, "moov", 4)) break;
    moov_offset += moov_len;
  }
  byte *moov = (byte*) malloc(moov_len);
  seek(f, moov_offset);
  fread(moov, 1, moov_len, f);
  
  // make sure we can decrypt the priv atom in each trak
  bool has_drm = false;
  {
    byte *trak = moov + 8;
    int n = moov_len - 8;
    while (n > 0) {
      if (!strncmp((char*)trak + 4, "trak", 4)) {
        if (decrypt_priv_in_trak(trak)) has_drm = true;
      }
      int trak_size = get_int32(trak);
      trak += trak_size;
      n -= trak_size;
    }
  }
  if (!has_drm) {
    // nothing to do - just leave original alone
    warning("no DRM found.\n");
    return 0;
  }
  
  // copy input file to output file unmodified.
  seek(f, 0);
  FILE *g = fopen(output_file, "wb");
  if (!g) error("can't open file %s\n", output_file);
  for (int i = 0; i < len; i += 4096) {
    int size = len - i < 4096 ? len - i : 4096;
    fread(buf, 1, size, f);
    fwrite(buf, 1, size, g);
  }
  
  // for each trak in the moov, decrypt it.
  byte *trak = moov + 8;
  int n = moov_len - 8;
  while (n > 0) {
    if (!strncmp((char*)trak + 4, "trak", 4)) {
      decrypt_trak(f, g, trak, moov_offset + (trak - moov));
    }
    int trak_size = get_int32(trak);
    trak += trak_size;
    n -= trak_size;
  }
  
  fclose(f);
  fclose(g);
  if (verbose) printf("decryption complete.\n");
  return 0;
}
